<?php
namespace Modules\FooBar\Crud;

use Illuminate\Support\Facades\Auth;
use Illuminate\Http\Request;
use App\Services\CrudService;
use App\Services\CrudEntry;
use App\Classes\CrudTable;
use App\Classes\CrudInput;
use App\Classes\CrudForm;
use App\Exceptions\NotAllowedException;
use TorMorten\Eventy\Facades\Events as Hook;
// use App\Modules\Brands;
use Modules\FooBar\Models\Brand; // Correct namespace


class BrandCrud extends CrudService
{
    /**
     * Defines if the crud class should be automatically discovered.
     * If set to "true", no need register that class on the "CrudServiceProvider".
     */
    const AUTOLOAD = true;

    /**
     * define the base table
     * @param string
     */
    protected $table = 'module_brands';

    /**
     * default slug
     * @param string
     */
    protected $slug = 'brands';

    /**
     * Define namespace
     * @param string
     */
    protected $namespace = 'foobar.brands';

    /**
     * To be able to autoload the class, we need to define
     * the identifier on a constant.
     */
    const IDENTIFIER = 'foobar.brands';

    /**
     * Model Used
     * @param string
     */
    protected $model = Brand::class;

    /**
     * Define permissions
     * @param array
     */
    protected $permissions = [
        'create' => true,
        'read' => true,
        'update' => true,
        'delete' => true,
    ];

    /**
     * Adding relation
     * Example : [ 'nexopos_users as user', 'user.id', '=', 'nexopos_orders.author' ]
     * @param array
     */
    // public $relations = [
    // ['nexopos_users as user', 'module_brands.created_by', '=', 'user.id']
    // ];

    public $relations = [
        // ['nexopos_users as user', 'module_brands.created_by', '=', 'user.id'],
        'leftJoin' => [
            ['nexopos_users as cuser', 'module_brands.created_by', '=', 'cuser.id'],
            ['nexopos_users as uuser', 'module_brands.updated_by', '=', 'uuser.id'],
        ],
    ];

    /**
     * all tabs mentionned on the tabs relations
     * are ignored on the parent model.
     */
    protected $tabsRelations = [
        // 'tab_name'      =>      [ YourRelatedModel::class, 'localkey_on_relatedmodel', 'foreignkey_on_crud_model' ],
    ];

    /**
     * Export Columns defines the columns that
     * should be included on the exported csv file.
     */
    protected $exportColumns = []; // @getColumns will be used by default.

    /**
     * Pick
     * Restrict columns you retrieve from relation.
     * Should be an array of associative keys, where
     * keys are either the related table or alias name.
     * Example : [
     *      'user'  =>  [ 'username' ], // here the relation on the table nexopos_users is using "user" as an alias
     * ]
     */
    public $pick = [
        // 'user' => ['first_name'],
        'cusers' => ['first_name'],
        'uusers' => ['first_name']
    ];

    /**
     * Define where statement
     * @var array
     **/
    protected $listWhere = [
        'status' => 1
    ];

    /**
     * Define where in statement
     * @var array
     */
    protected $whereIn = [];

    /**
     * If few fields should only be filled
     * those should be listed here.
     */
    public $fillable = ['name', 'logo', 'description', 'website', 'email', 'phone_number', 'address', 'created_by', 'updated_by'];

    /**
     * If fields should be ignored during saving
     * those fields should be listed here
     */
    public $skippable = [];

    /**
     * Determine if the options column should display
     * before the crud columns
     */
    protected $prependOptions = false;

    /**
     * Will make the options column available per row if
     * set to "true". Otherwise it will be hidden.
     */
    protected $showOptions = true;

    /**
     * In case this crud instance is used on a search-select field,
     * the following attributes are used to auto-populate the "options" attribute.
     */
    protected $optionAttribute = [
        'value' => 'id',
        'label' => 'name'
    ];

    /**
     * Return the label used for the crud object.
     **/
    public function getLabels(): array
    {
        return CrudTable::labels(
            list_title: __('Brands List'),
            list_description: __('Display all brands.'),
            no_entry: __('No brands has been registered'),
            create_new: __('Add a new brand'),
            create_title: __('Create a new brand'),
            create_description: __('Register a new brand and save it.'),
            edit_title: __('Edit brand'),
            edit_description: __('Modify  Brand.'),
            back_to_list: __('Return to Brands'),
        );
    }

    /**
     * Defines the forms used to create and update entries.
     * @param Brand $entry
     * @return array
     */
    public function getForm(Brand $entry = null): array
    {
        return CrudForm::form(
            main: CrudInput::text(
                label: __('Name'),
                name: 'name',
                validation: 'required',
                description: __('Provide a name to the resource.'),
                value: $entry->value ?? '',
            ),
            tabs: CrudForm::tabs(
                CrudForm::tab(
                    identifier: 'general',
                    label: __('Information'),
                    fields: CrudForm::fields(
                        CrudInput::text(
                            type: 'media',
                            label: __('Logo'),
                            name: 'logo',
                            value: $entry->value ?? '',
                            description: __('Provide a name to the resource.'),
                        ),
                        CrudInput::text(
                            label: __('Website'),
                            name: 'website',
                            value: $entry->value ?? '',
                            description: __('Provide a name to the resource.'),
                        ),
                        CrudInput::text(
                            label: __('Email'),
                            name: 'email',
                            value: $entry->value ?? '',
                            description: __('Provide a name to the resource.'),
                        ),
                        CrudInput::text(
                            label: __('Phone Number'),
                            name: 'phone_number',
                            value: $entry->value ?? '',
                            description: __('Provide a name to the resource.'),
                        ),
                        CrudInput::text(
                            label: __('Address'),
                            name: 'address',
                            value: $entry->value ?? '',
                            description: __('Provide a name to the resource.'),
                        ),
                        CrudInput::text(
                            label: __('Description'),
                            type: 'textarea',
                            name: 'description',
                            value: $entry->value ?? '',
                            description: __('Provide a name to the resource.'),
                        ),
                    )
                )
            )
        );
    }

    /**
     * Filter POST input fields
     * @param array of fields
     * @return array of fields
     */
    public function filterPostInputs($inputs): array
    {
        foreach ($inputs as $key => $entry) {
            $inputs[$key] = $entry ?? '';
        }
        // Check if the user is authenticated
        if (Auth::check()) {
            // Assign the authenticated user's ID to created_by
            $inputs['created_by'] = Auth::id();
            $inputs['updated_by'] = null;
        } else {
            // Log if the user is not authenticated
            \Log::warning('User is not authenticated when trying to create a record.');
        }

        \Log::info('1 Before Post:', $inputs); // Log the request data
        return $inputs;
    }

    /**
     * Filter PUT input fields
     * @param array of fields
     * @return array of fields
     */
    public function filterPutInputs(array $inputs, Brand $entry)
    {
        return $inputs;
    }

    /**
     * Trigger actions that are executed before the
     * crud entry is created.
     */
    public function beforePost(array $request): array
    {
        // Check if the user is allowed to create
        $this->allowedTo('create');

        return $request;
    }

    /**
     * Trigger actions that will be executed 
     * after the entry has been created.
     */
    public function afterPost(array $request, Brand $entry): array
    {
        return $request;
    }


    /**
     * A shortcut and secure way to access
     * senstive value on a read only way.
     */
    public function get(string $param): mixed
    {
        switch ($param) {
            case 'model':
                return $this->model;
                break;
        }
    }

    /**
     * Trigger actions that are executed before
     * the crud entry is updated.
     */
    public function beforePut(array $request, Brand $entry): array
    {
        $this->allowedTo('update');

        return $request;
    }

    /**
     * This trigger actions that are executed after
     * the crud entry is successfully updated.
     */
    public function afterPut(array $request, Brand $entry): array
    {
        return $request;
    }

    /**
     * This triggers actions that will be executed ebfore
     * the crud entry is deleted.
     */
    public function beforeDelete($namespace, $id, $model): void
    {
        if ($namespace == 'foobar.brands') {
            /**
             *  Perform an action before deleting an entry
             *  In case something wrong, this response can be returned
             *
             *  return response([
             *      'status'    =>  'danger',
             *      'message'   =>  __( 'You\re not allowed to do that.' )
             *  ], 403 );
             **/
            if ($this->permissions['delete'] !== false) {
                ns()->restrict($this->permissions['delete']);
            } else {
                throw new NotAllowedException;
            }
        }
    }

    /**
     * Define columns and how it is structured.
     */
    public function getColumns(): array
    {
        return CrudTable::columns(
            CrudTable::column(
                identifier: 'name',
                label: __('Name'),
            ),
            CrudTable::column(
                identifier: 'logo',
                label: __('Logo'),
            ),
            CrudTable::column(
                identifier: 'description',
                label: __('Description'),
            ),
            CrudTable::column(
                identifier: 'website',
                label: __('Website'),
            ),
            CrudTable::column(
                identifier: 'email',
                label: __('Email'),
            ),
            CrudTable::column(
                identifier: 'phone_number',
                label: __('Phone Number'),
            ),
            CrudTable::column(
                identifier: 'address',
                label: __('Address'),
            ),
            CrudTable::column(
                identifier: 'cuser_first_name',
                label: __('Created By'),
            ),
            CrudTable::column(
                identifier: 'uuser_first_name',
                label: __('Updated By'),
            ),
            CrudTable::column(
                identifier: 'created_at',
                label: __('Created At'),
            ),
            CrudTable::column(
                identifier: 'updated_at',
                label: __('Updated At'),
            ),
        );
    }

    /**
     * Define row actions.
     */
    public function setActions(CrudEntry $entry): CrudEntry
    {

        //TODO: Fix this
        /**
         * Declaring entry actions
         */
        // $entry->action(
        //     identifier: 'edit',
        //     label: __('Edit'),
        //     url: ns()->url('/dashboard/' . $this->slug . '/edit/' . $entry->id)
        // );

        $entry->action(
            identifier: 'delete',
            label: __('Delete'),
            type: 'DELETE',
            url: ns()->url('/api/crud/foobar.brands/' . $entry->id),
            confirm: [
                'message' => __('Would you like to delete this ?'),
            ]
        );

        return $entry;
    }


    /**
     * trigger actions that are executed
     * when a bulk actio is posted.
     */
    public function bulkAction(Request $request): array
    {
        /**
         * Deleting licence is only allowed for admin
         * and supervisor.
         */

        if ($request->input('action') == 'delete_selected') {

            /**
             * Will control if the user has the permissoin to do that.
             */
            if ($this->permissions['delete'] !== false) {
                ns()->restrict($this->permissions['delete']);
            } else {
                throw new NotAllowedException;
            }

            $status = [
                'success' => 0,
                'error' => 0
            ];

            foreach ($request->input('entries') as $id) {
                $entity = $this->model::find($id);
                if ($entity instanceof Brands) {
                    $entity->delete();
                    $status['success']++;
                } else {
                    $status['error']++;
                }
            }
            return $status;
        }

        return Hook::filter($this->namespace . '-catch-action', false, $request);
    }

    /**
     * Defines links used on the CRUD object.
     */
    public function getLinks(): array
    {
        return CrudTable::links(
            list: ns()->url('dashboard/' . 'brands'),
            create: ns()->url('dashboard/' . 'brands/create'),
            edit: ns()->url('dashboard/' . 'brands/edit/'),
            post: ns()->url('api/crud/' . 'foobar.brands'),
            put: ns()->url('api/crud/' . 'foobar.brand/{id}' . ''),
        );
    }

    /**
     * Defines the bulk actions.
     **/
    public function getBulkActions(): array
    {
        return Hook::filter($this->namespace . '-bulk', [
            [
                'label' => __('Delete Selected Entries'),
                'identifier' => 'delete_selected',
                'url' => ns()->route('ns.api.crud-bulk-actions', [
                    'namespace' => $this->namespace
                ])
            ]
        ]);
    }

    /**
     * Defines the export configuration.
     **/
    public function getExports(): array
    {
        return [];
    }
}
